; Boot-Man
;
; (c) 2019 Guido van den Heuvel
;
; Boot-Man is a Pac-Man clone that fits (snugly) inside the Master Boot Record of a USB stick.
; A USB stick with Boot-Man on it boots into the game (hence the name). Unfortunately, however,
; Boot-Man leaves no room in the MBR for a partition table, which means that a USB stick with Boot-Man
; in its MBR cannot be used to store data. In fact, Windows does not recognize a USB stick with 
; Boot-Man in its MBR as a valid storage medium.
;
; Controls of the game: you control Boot-Man using the arrow keys. No other user input is necessary. Some other
; keys can also be used to control Boot-Man, this is a side effect of coding my own keyboard handler in 
; just a few bytes. There simply wasn't room for checking the validity of every key press.
;
; The game starts automatically, and when Boot-Man dies, a new game starts automatically after any key hit.
;
;
; I've had to take a couple of liberties with the original Pac-Man game to fit Boot-Man inside the 510
; bytes available in the MBR:
;
; * The ghosts start in the four corners of the maze, they do not emerge from a central cage like in the original
;
; * There's just a single level. If you finish the level, the game keeps running with an empty maze. While 
;   it is rather difficult to finish the game (which is intentional, because of the single level), it is possible. 
;
; * Boot-Man only has 1 life. If Boot-Man dies, another game is started automatically.
;
; * Power pills function differently from the original. When Boot-Man eats a power pill, all ghosts become
;   ethereal (represented in game by just their eyes being visible) and cease to chase Boot-Man. While ethereal,
;   Boot-Man can walk through ghosts with no ill effects. While I would really like to include the "ghost hunting"
;   from the original, which I consider to be an iconic part of the game, this simply isn't possible in the little
;   space available.
;
; * There's no score, and no fruit to get bonus points from.
; 
; * All ghosts, as well as Boot-Man itself, have the same, constant movement speed. In the original, the ghosts
;   run at higher speeds than Pac-Man, while Pac-Man gets delayed slightly when eating and ghosts get delayed when moving
;   through the tunnel connecting both sides of the maze. This leads to very interesting dynamics and strategies
;   in the original that Boot-Man, by necessity, lacks.
;
;
; Boot-Man runs in text mode. It uses some of the graphical characters found in IBM codepage 437 for its objects:
;   - Boot-Man itself is represented by the smiley face (☻), which is character 0x02 in the IBM charset
;   - The Ghosts are represented by the infinity symbol (∞), which is character 0xec. These represent
;     a ghost's eyes, with the ghost's body being represented simply by putting the character on a 
;     coloured background
;   - The dots that represent Boot-Man's food are represented by the bullet character (•), 
;     which is character 0xf9
;   - The power pills with which Boot-Man gains extra powers are represented by the diamond (♦),
;     which is character 0x04
;   - The walls of the maze are represented by the full block character (█), which is character 0xdb
;
; Boot-Man runs off BIOS clock. It should therefore run at the same speed on all PCs. The code is quite heavily
;  optimized for size, so code quality is questionable at best, and downright atrocious at worst.

;%define WaitForAnyKey               ; +9 bytes
%define MDAsupport                  ; +7 bytes
%define MDAinverse                  ; +5 bytes
%define MDA_CGA40                   ; +17 bytes

cpu 8086                            ; Boot-Man runs on the original IBM PC with a CGA card.
bits 16                             ; Boot-Man runs in Real Mode. I am assuming that the BIOS leaves the CPU is Real Mode.
                                    ; This is true for the vast majority of PC systems. If your system's BIOS
                                    ; switches to Protected Mode or Long Mode during the boot process, Boot-Man
                                    ; won't run on your machine.
org 0x600

%define COL40         40
%define COL80         80
   
start:
    call copy                       ; Can run as bootsector or DOS COM file
    
moved:
    push cs
    pop ss
    xor sp, sp                      ; Set up the stack.

    mov ax, 1                       ; int 0x10 / ah = 0: Switch video mode. Switch mode to 40x25 characters (al = 1).
    int 0x10                        ; In this mode, characters are approximately square, which means that horizontal 
                                    ; and vertical movement speeds are almost the same.
    mov ax, 0xb800
%ifdef MDAsupport
 %define LeftCorner    (COL80*2*24+COL80-32)
 %define PreviousLine  (COL80*2+32)
 %define Columns       COL80
 %define Margin        ((COL80-32)/2)
 %ifdef MDA_CGA40
    mov dx, PreviousLine
   %undef PreviousLine
   %define PreviousLine dx
 %endif
%else
 %define LeftCorner    (COL40*2*24+COL40-32)
 %define PreviousLine  (COL40*2+32)
 %define Columns       COL40
 %define Margin        ((COL40-32)/2)
%endif
initES:
    mov es, ax                      ; Set up the es segment to point to video RAM
    mov si, maze                    ; The first byte of the bit array containing the maze
%define ghost_n_incr 0x2f10                                                       
%ifdef MDAsupport
    mov ah, 0xb0
 %ifdef MDAinverse
   %define ghost_attr 0x70
   %define incr_attr  0x80
   %undef ghost_n_incr
   %define ghost_n_incr (ghost_attr*256 + incr_attr)
    xor word [si + ghost_attr_patch - maze], ghost_n_incr^0x2f10
 %endif
 %ifdef MDA_CGA40
;    xor dl, (COL80*2+32)^(COL40*2+32)
    xor dl, ah
    xor word [si + LeftCorner_patch - maze], LeftCorner^(COL40*2*24+COL40-32)
    xor byte [si + Columns_patch - maze], Columns^COL40
    xor byte [si + Margin_patch - maze], Margin^((COL40-32)/2)
 %endif
    scasw
    jb initES
%endif

    mov ax, 0x0101                  ; int 0x10 / ah = 1: Determine shape of hardware cursor. al = 1 avoid AMI BIOS bug.
    mov ch, 0x20                    ; With cx = 0x2000, this removes the hardware cursor from the screen altogether
    int 0x10

%define ghost_color        0x2fec
%define color_increment    0x10


;-----------------------------------------------------------------------------------------------------
; buildmaze: builds the maze. The maze is stored in memory as a bit array, with 1 representing a wall
;            and 0 representing a food dot. Since the maze is left-right symmetrical, only half of the
;            maze is stored in memory. The positions of the power pills is hard-coded in the code.
;            Adding the power pills to the bit array would have necessitated 2 bits for every 
;            character, increasing its size drastically.
;
;            Both sides of the maze are drawn simultaneously. The left part is drawn left to right,
;            while the right part is drawn right to left. For efficiency reasons, the entire maze 
;            is built from the bottom up. Therefore, the maze is stored upside down in memory
;-----------------------------------------------------------------------------------------------------
;buildmaze:
    mov di, LeftCorner              ; Lower left corner of maze in video ram
LeftCorner_patch      equ $ - 2
    mov cx, 34                      ; Lines count to the lower left powerpill
.maze_outerloop:
    mov bx, 0x003c                  ; The distance between a character in the maze and its 
                                    ; symmetric counterpart. Also functions as loop counter
    lodsw                           ; Read 16 bits from the bit array, which represents one
    xchg ax, bp                     ; 32 character-wide row of the maze
.maze_innerloop:
    add bp, bp                      ; shift out a single bit to determine whether a wall or dot must be shown
    mov ax, 0x01db                  ; Assume it is a wall character (0x01: blue; 0xdb: full solid block)
    jc .draw                        ; Draw the character if a 1 was shifted out
    mov ax, 0x0ff9                  ; otherwise, assume a food character (0x0f: white; x0f9: bullet)
    loop .draw                      ; See if instead of food we need to draw a power pill
    mov cl, 125                     ; Update powerpill address to draw remaining powerpills
    mov al, 0x04                    ; powerpill character (0x04: diamond - no need to set up colour again)
.draw:
    stosw                           ; Store character + colour in video ram
    mov [es:bx+di], ax              ; Go to its symmetric counterpart and store it as well
    sub bx, 4                       ; Update the distance between the two sides of the maze
    jns .maze_innerloop             ; As long as the distance between the two halves is positive, we continue

    sub di, PreviousLine            ; Go to the previous line on the screen in video RAM. 
    jns .maze_outerloop             ; Keep going as long as this line is on screen.

    ;mov si, bootman_data            ; Use si as a pointer to the game data. This reduces byte count of the code:
                                    ; mov reg, [address] is a 4 byte instruction, while mov reg, [si] only has 2.

;-----------------------------------------------------------------------------------------------------
; game_loop:   The main loop of the game. Tied to BIOS clock (which fires 18x per 
;              second by default), to ensure that the game runs at the same speed on all machines
;
;              The code first updates Boot-Man's position according to its movement direction
;              and keyboard input. Then the ghost AI is run, to determine the ghosts' movement
;              direction and update their position. Finally, Boot-Man and the ghosts are drawn
;              in their new positions. Collisions between Boot-Man and the ghosts are checked
;              before and after ghost movement. We need to detect for collisions twice, because
;              if you only check once, Boot-Man can change position with a ghost without colliding
;              (in the original, collisions are checked only once, and as a consequence, it is 
;              possible in some circumstances to actually move Pac-Man through a ghost). 
;-----------------------------------------------------------------------------------------------------

game_loop:
    hlt                             ; Wait for clock interrupt
    mov ah, 0x00
    int 0x1a                        ; BIOS clock read
    xchg ax, dx
old_time equ $ + 1
    cmp al, 0x12                    ; Wait for time change
    je game_loop
    mov [old_time], al              ; Save new time

    mov bx, 3
    mov ah, 0x01                    ; BIOS Key available
    int 0x16
    cbw                             ; BIOS Read Key
    je .no_key
    int 0x16
    mov al, ah
    dec ah                          ; Escape ?
    jne .convert_scancode
;.exit:
    xchg ax, bx                     ; int 0x10 / ax = 3: Switch video mode. Switch mode to 80x25 characters
    int 0x10                        ; Reset screen
    int 0x20                        ; Exit to DOS...
    int 0x19                        ; ...or reboot

    ; This code converts al from scancode to movement direction.
    ; Input:  0x48 (up), 0x4b (right), 0x50 (down), 0x4d (left)
    ; Output: 0xce (up), 0xca (right), 0xc6 (down), 0xc2 (left) 
    ; fe xx:  dec dh     dec dl        inc dh       inc dl
    
.convert_scancode:
    sub al, 0x47                    ; 0x01 0x04 0x09 0x06
    and al, 0xfe                    ; 0x00 0x04 0x08 0x06
    cmp al, 9
    jnc .no_key                     ;                      if al >= 0x50, ignore scancode;
                                    ;                      this includes key release events
    neg al                          ; 0x00 0xfc 0xf8 0xfa
    add al, 0xce                    ; 0xce 0xca 0xc6 0xc8
    test al, 2
    jne .translated
    mov al, 0xc2
.translated:                       
    cmp al, [si + 2]                ; If the new direction is the same as the current direction, ignore it
    jz .no_key
    mov [bx+si], al                 ; Set new direction to the direction derived from the keyboard input
.no_key:
    dec byte [si + pace_offset]     ; Decrease the pace counter. The pace counter determines the overall
                                    ; speed of the game. We found that moving at a speed of 6 per second
                                    ; gives good speed and control, so we use a counter to only move
                                    ; once for every three times that the interrupt fires.
                                    ; We also use the pace counter to include longer delays at game start
                                    ; and after Boot-Man dies, by intitalizing the counter with higher values.
    jnz game_loop                   ; If the pace counter is not 0, we immediately finish.

    mov byte [si + pace_offset], bl ; Reset the pace counter to 3.
;-----------------------------------------------------------------------------------------------------
; Move Boot-Man
;-----------------------------------------------------------------------------------------------------
    call newpos_bootman             ; Update dx to move 1 square in the direction indicated by al
                                    ; newpos also checks for collisions with walls (in which case ZF is set)
    jz .nodirchange                 ; ZF indicates that new position collides with wall. We therefore try to keep
                                    ; moving in the current direction instead.
    mov [bx+si], al                 ; If there's no collision, update the current movement direction
.nodirchange:
    call newpos_bootman             ; Update dx to move 1 square in direction al
    jz .endbootman                  ; If there's a wall there, do nothing
;.move:
    mov ax, 0x0f20                  ; Prepare to remove Boot-Man from screen, before drawing it in the new position
                                    ; 0x0f = black background, white foreground; 0x20 = space character
    cmp byte [es:di], ah            ; Detect power pill (0x04)
    ja .nopowerpill                
    mov byte [si + timer_offset], al; If Boot-Man just ate a power pill, set up the ghost timer to 0x20. We use al here
                                    ; as it accidentally contains 0x20, and this is one byte shorter than having an 
                                    ; explicit constant.
.nopowerpill:
    xchg dx, [si + bm_offset_pos]   ; Retrieve current position and store new position
    call paint                      ; Actually remove Boot-Man from the screen
.endbootman:
;-----------------------------------------------------------------------------------------------------
; ghost AI and movement
; 
; Determine the new movement direction for each ghost. Ghost movement direction is determined by
; the following rule:
; (1) Every ghost must keep moving
; (2) It is forbidden for ghosts to suddenly start moving backwards. Unless Boot-Man just consumed 
;     a powerpill, in which case ghosts are forbidden from continuing in the direction they were going
; (3) Whenever a ghost has multiple movement options (i.e., it is at a crossroads), try moving 1 space
;     in each direction that is allowed, and calculate the distance to the target location after 
;     that move. Choose the direction for which this distance is lowest as the new movement direction
;
; During normal movement, ghosts target a position that is related to the position of Boot-Man, as follows:
; 
; number | ghost colour | target
; -------+--------------+-------------------
;      1 | purple       | bootman's position
;      2 | red          | 4 squares below Boot-Man
;      3 | cyan         | 4 squares to the left of Boot-Man
;      4 | green        | 4 squares to the right of Boot-Man
;
; There's two different reasons for having slightly different AI for each ghost:
; (1) If all ghosts have the same AI they tend to bunch together and stay there. With the current AI, 
;     ghosts will sometimes bunch together, but they will split apart eventually
; (2) With this setup, the ghosts tend to surround Boot-Man, making it harder for the player
; 
; When Boot-Man picks up a power pill, a timer starts running, and ghosts become ethereal.
; As long as the ghosts are ethereal, the 
; ghosts will not chase Boot-Man. Instead they will use the center of the big rectangular block
; in the middle of the maze as their target. They cannot reach it, obviously, so the result is 
; that they will keep circling this block for as long as the timer runs.
;
; This AI is related to, but not the same as, the AI actually used in Pac-Man. The red Pac-Man ghost
; uses Pac-Man itself as target, same as my purple ghost, while the pink Pac-Man ghost will 
; target 4 squares ahead of Pac-Man, in the direction Pac-Man is currently moving. The other ghosts' 
; movement is a bit more complex than that. I had to simplify the AI because of the limited code size.
;-----------------------------------------------------------------------------------------------------
    mov bl, 3 * gh_length + bm_length       ; Set up offset to ghost data. With this, si + bx is a 
                                            ; pointer to the data from the last ghost. Also used as 
                                            ; loop counter to loop through all the ghosts  
.ghost_ai_outer:
    mov bp, si                              ; bp = minimum distance; start out at a big int
    mov ah, [bx + si + gh_offset_dir]       ; ah will become the forbidden movement direction. We start
                                            ; with the current direction, which is forbidden if Boot-Man
                                            ; just ate a power pill
    cmp byte [si + timer_offset], 0x20      ; If timer_offset == 0x20, Boot-Man just picked up a power pill
    jz .reverse                             ; so in that case we do not flip the direction.
    xor ah, 8                               ; Flip the current direction to obtain the forbidden direction in ah
.reverse:
    mov al, 0xce                            ; al = current directions being tried. Doubles as loop counter
                                            ; over all directions.
                                            ; Values are the same as those used by the newpos routine
    call test_collision                     ; dx = current ghost position
.ghost_ai_loop:
    push dx
    cmp al, ah                              ; If the current direction is the forbidden direction
    jz .next                                ; we continue with the next direction
    call newpos                             ; Update ghost position and check if it collides with a wall
    jz .next                                ; if so, we continue with the next direction
    push ax
    mov ax, 0x0c10                          ; Target position if ghosts are ethereal. Position 0x0c10 
                                            ; (x = 0x10, y = 0x0c) is in the center of the maze.
    cmp byte [si + timer_offset], bh        ; See if ghost timer runs. We compare with bh, which is known to be 0.
    jnz .skip_target                        ; If ghost timer runs, we use the aforementioned target position
    mov ax, [si + bm_offset_pos]            ; Otherwise we use Boot-Man's current position,
    add ax, [bx + si + gh_offset_focus]     ; Updated with an offset that is different for each ghost
.skip_target:
;-----------------------------------------------------------------------------------------------------
; get_distance: Calculate distance between positions in cx (target position) and dx (ghost position)
;               This used to be a function, but I inlined it to save some space.
;               The square of the distance between the positions in cx and dx is calculated,
;               according to Pythagoras' theorem.
;-----------------------------------------------------------------------------------------------------
    sub al, dl                              ; after this, al contains the horizontal difference
    sub ah, dh                              ; and ah the vertical difference

    mov cl, ah       
    imul al
    xchg ax, cx                             ; cx = square of horizontal difference
    imul al                                 ; ax = square of vertical difference
    add cx, ax                              ; cx = distance squared between positions in cx and dx
    pop ax

    cmp cx, bp                              ; Compare this distance to the current minimum
    jnc .next                               ; and if it is,
    mov bp, cx                              ; update the minimum distance
    mov [bx + si + gh_offset_dir], al       ; set the movement direction to the current direction
    mov [bx + si + gh_offset_pos], dx       ; Store the new ghost position
.next:
    pop dx                                  ; Restore the current ghost position 
    sub al, 4                               ; Update the current direction / loop counter
    cmp al, 0xc2                            
    jnc .ghost_ai_loop

    mov ax, [bx + si + gh_offset_terrain]   ; Remove the ghost in the old position from the screen 
    call paint                              ; by painting the terrain underneath that ghost that was 
                                            ; determined in the previous movement phase.
    sub bx, gh_length                       ; Go to the next ghost,
    jns .ghost_ai_outer                     ; and stop after the final ghost


.ghostterrain_loop:                         ; Second loop through all the ghosts, to determine terrain
                                            ; underneath each one. This is used in the next movement phase
                                            ; to restore the terrain underneath the ghosts.
                                            ; Note that this "terrain storing" approach can trigger a bug
                                            ; if Boot-Man and a ghost share a position. In that case
                                            ; an extra Boot-Man character may be drawn on screen.
    add bx, gh_length                       ; go to next ghost
    call test_collision                     ; dx = current ghost position
    call get_screenpos                      ; find the address in video ram of the updated ghost position,
    mov ax, [es:di]                         ; store its content in ax
    mov [bx + si + gh_offset_terrain], ax   ; and copy it to ghostterrain
    cmp bx, 3 * gh_length + bm_length       ; and determine if it is the final ghost
    jnz .ghostterrain_loop

    mov ax, 0x0f00                          ; Update ghost colour to black background, white eyes
                                            ; Update difference between colours of successive ghosts. Value of 0x0
                                            ; means all ghosts are the same colour when they are ethereal.
    dec byte [si + timer_offset]            ; Update ghost_timer to limit the period of the ghosts being ethereal
    jns .ghostcolor
    mov byte [si + timer_offset], bh        ; Ghost timer was not running
    mov ax, ghost_n_incr
ghost_attr_patch  equ $ - 2
 
.ghostcolor:
    mov cl, al                              ; cl = difference in colour between successive ghosts
    mov al, ghost_color&255                 ; 0xec = infinity symbol = ghost eyes 
.ghostdraw:                                 ; Draw the ghosts on the screen
    mov dx, [bx + si + gh_offset_pos]       ; dx = new ghost position
    call paint                              ; show ghost in video ram
    add ah, cl                              ; Update ghost colour.
    sub bl, gh_length                       ; Loop over all ghosts
    jns .ghostdraw                          ; until the final one.

    mov ax, game_loop                       ; ret instruction will jump to game_loop
    push ax
    mov ax, word 0x0e02                     ; Draw boot-man on the screen. 0x0e = black background, yellow foreground
                                            ; 0x02 = smiley face
;-----------------------------------------------------------------------------------------------------
; paint: paints a character on screen at given x, y coordinates in dx
;        simple convenience function that gets called enough to be actually worth it, in terms
;        of code length.
;-----------------------------------------------------------------------------------------------------
paint_bootman:
    mov dx, [si + bm_offset_pos]            ; dx = Boot-Man position
paint:
    call get_screenpos                      ; Convert x, y coordinates in dx to video memory address
    stosw                                   ; stosw = shorter code for mov [es:di], ax
                                            ; stosw also adds 2 to di, but that effect is ignored here
    ret


;-----------------------------------------------------------------------------------------------------
; test_collision: if end of game, restart from the beginning
;-----------------------------------------------------------------------------------------------------

test_collision:
    mov dx, [bx + si + gh_offset_pos]       ; dx = current ghost position
    cmp dx, [si + bm_offset_pos]            ; compare dx with Boot-Man position
    jnz no_collision
    cmp byte [si + timer_offset], bh        ; Ghost timer was not running
    jnz no_collision
    pop ax                                  ; Adjust stack
%ifdef WaitForAnyKey
    ; Ghosts are visible and collide with boot-man, therefore boot-man is dead
    mov ax, 0x0e0f                          ; Dead boot-man: 0x0e = black background, yellow foreground
                                            ; 0x0f = 8 pointed star
    call paint_bootman
    cbw
    int 0x16                                ; Wait for any key
%endif
    mov si, 0xfb5e                          ; restart boot sector
source equ $ - 2


;-----------------------------------------------------------------------------------------------------
; copy: self copy to a fixed location
;-----------------------------------------------------------------------------------------------------

copy equ $ - 2                              ; seek to 'pop si, sti' instruction (0x5e 0xfb)
;    pop si                                  ; Get ip value (0x103 or 0x7C03, sometimes 0x0003)
;    sti                                     ; Allow interrupts
    push cs
    pop ds
    push cs                                 ; Setup ds and es
    pop es
    mov di, moved                           ; Move self to a well known address
    push di
    mov ch, (end-moved)/256+1
    cld                                     ; Clear the direction flag. We use string instructions a lot as they have one-byte codes
    mov [si+source-moved], si               ; Save value to the restart with unpatched code 
    rep movsb
no_collision:
    ret


;-----------------------------------------------------------------------------------------------------
; newpos: calculates a new position, starting from a position in dx and movement direction in al.
;         dl contains the x coordinate, while dh contains the y coordinate. The movement directions
;         in al are as follows:
;         0xc2: move right
;         0xc6: move down
;         0xca: move left
;         0xce: move up
;
; The reason for these fairly strange values is that they form the 2nd byte (the ModR/M byte)
; of the instruction updating the position: 
; inc dl (0xfe, 0xc2), inc dh (0xfe), dec dl (0xfe, 0xca), dec dh (0xfe, 0xce)
; The code first modifies itself to the correct instruction, then executes this instruction. The
; reason for doing it in this way is that this is a lot shorter than the traditional way of 
; doing an if / elif / elif / else chain.
;
; Immediately after calculating the new position we also determine the address in video RAM 
; corresponding to this position. All lines of the screen are stored one after the other in RAM,
; starting at 0xb800:0x0000. Since each line has 40 characters, and every character takes up
; two bytes (one for colour, one for the character code), the equation to calculate video RAM
; offset from x, y coordinates is as follows:
; 
; offset = 2 * (40 * y + x + 4),
;
; with the +4 due to the fact that the maze is in the center of the screen, with a 4 character wide
; border to the left.
;
; newpos and get_screenpos used to be two separate functions but since they were almost
; always called one after the other, combining them saved some bytes of code.
;-----------------------------------------------------------------------------------------------------

newpos_bootman:
    mov al, [bx+si]                         ; al = new or current movement direction
    dec bx
    mov dx, [si + bm_offset_pos]            ; dx = current position of Boot-Man
newpos:
    mov [.modified_instruction + 1], al     ; Here the instruction to be executed is modified
.modified_instruction:
    db 0xfe, 0xc2                           ; inc dl in machine code
    and dl, 0x1f                            ; Deal with tunnels
get_screenpos:
    xchg ax,di                              ; save ax
    mov al, Columns
Columns_patch         equ $ - 1
    mul dh                                  ; multiply ax by 0x28 = 40 decimal, the screen width
    add al, dl
    adc ah, bh
    xchg ax, di                             ; di = y * 40 + x
%ifdef MDAsupport
    add di, Margin
Margin_patch          equ $ - 1
%else
    scasw                                   ; Skip the left border by adding 4 to di
    scasw
%endif
    add di, di                              ; Multiply di by 2
    cmp byte [es:di], 0xdb                  ; Check to see if the new position collides with a wall
                                            ; 0xdb = full block character that makes up the wall
    ret

; The maze, as a bit array. Ones denote walls, zeroes denote food dots / corridors
; The maze is stored upside down to save one cmp instruction in buildmaze
maze: dw 0xffff, 0x8000, 0xbffd, 0x8081, 0xfabf, 0x8200, 0xbefd, 0x8001
      dw 0xfebf, 0x0080, 0xfebf, 0x803f, 0xaebf, 0xaebf, 0x80bf, 0xfebf
      dw 0x0080, 0xfefd, 0x8081, 0xbebf, 0x8000, 0xbefd, 0xbefd, 0x8001
      dw 0xffff

bootman_data:
bootmanpos:
    db 0x0f, 0x0f               ; Boot-Man's x and y position
bootmandir:
    db 0xca                     ; Boot-Man's direction
    db 0xca                     ; Boot-Man's future direction

ghost_timer      equ maze + 2   ; if > 0 ghosts are invisible, and is counted backwards to 0
;pace_counter     equ maze + 22  ; 0x3f
pace_counter: db 0x10

ghostdata:
ghostpos:
    db 0x01, 0x01               ; 1st ghost, x and y position
ghostdir:
    db 0xc2                     ;            direction
ghostterrain:
    dw 0x0ff9                   ;            terrain underneath
ghostfocus:
    db 0x0, 0x0                 ;            focus point for movement
secondghost:
    db 0x01, 0x17               ; 2nd ghost, x and y position
    db 0xce                     ;            direction
    dw 0x0ff9                   ;            terrain underneath
    db 0x0, 0x4                 ;            focus point for movement
    db 0x1e, 0x01               ; 3rd ghost, x and y position
    db 0xca                     ;            direction
    dw 0x0ff9                   ;            terrain underneath
    db 0xfc, 0x0                ;            focus point for movement
    db 0x1e, 0x17               ; 4th ghost, x and y position
    db 0xce                     ;            direction
    dw 0x0ff9                   ;            terrain underneath
    db 0x4, 0x0                 ;            focus point for movement
lastghost:

bm_length           equ ghostdata    - bootman_data
gh_length           equ secondghost  - ghostdata
gh_offset_dir       equ ghostdir     - ghostdata
gh_offset_pos       equ ghostpos     - ghostdata
gh_offset_terrain   equ ghostterrain - ghostdata
gh_offset_focus     equ ghostfocus   - ghostdata
pace_offset         equ pace_counter - bootman_data
timer_offset        equ ghost_timer  - bootman_data
bm_offset_pos       equ bootmanpos   - bootman_data

times 510 - ($ - $$) db 0
db 0x55
db 0xaa

end:
